@@ -1,20 +1,23 @@
 import re
 import os
-import Levenshtein
 from pathlib import Path
 from datetime import datetime
-from functools import lru_cache, cache
-from normality.constants import WS
 from followthemoney import model
+from functools import lru_cache, cache
+from jellyfish import damerau_levenshtein_distance, metaphone
+from jellyfish import jaro_winkler_similarity, soundex
+from normality import collapse_spaces
+from normality.constants import WS
 from collections.abc import Mapping, Sequence
-from fingerprints.fingerprint import fingerprint
-from fingerprints.cleanup import clean_strict
+from fingerprints.cleanup import clean_name_ascii, clean_entity_prefix
+from fingerprints import replace_types
 from followthemoney.util import sanitize_text
-from typing import cast, Any, Union, Iterable, Tuple, Optional, List, Set
+from typing import cast, Any, Union, Iterable, Tuple, Optional, List, Callable
 from urllib.parse import urlparse, urlunparse, parse_qsl, urlencode
 
 DATA_PATH = Path(os.path.join(os.path.dirname(__file__), "data")).resolve()
 QID = re.compile(r"^Q(\d+)$")
+ID_CLEAN = re.compile(r"[^A-Z0-9]+", re.UNICODE)
 BASE_ID = "id"
 PathLike = Union[str, os.PathLike[str]]
 ParamsType = Union[None, Iterable[Tuple[str, Any]], Mapping[str, Any]]
@@ -109,31 +112,93 @@ def text_bool(text: Optional[str]) -> Optional[bool]:
 
 
 @lru_cache(maxsize=1024)
-def fingerprint_name(original: str, keep_order: bool = True) -> Optional[str]:
+def fingerprint_name(original: str) -> Optional[str]:
     """Fingerprint a legal entity name."""
-    return fingerprint(original, keep_order=keep_order, keep_brackets=True)
+    # this needs to happen before the replacements
+    text = original.lower()
+    text = clean_entity_prefix(text)
+    # Super hard-core string scrubbing
+    cleaned = clean_name_ascii(text)
+    cleaned = replace_types(cleaned)
+    return collapse_spaces(cleaned)
 
 
-def name_words(names: List[str]) -> Set[str]:
-    """Get a unique set of tokens present in the given set of names."""
-    words: Set[str] = set()
+def names_word_list(
+    names: Iterable[str],
+    normalizer: Callable[[str], Optional[str]] = fingerprint_name,
+    processor: Optional[Callable[[str], str]] = None,
+    min_length: int = 1,
+) -> List[str]:
+    """Get a list of tokens present in the given set of names."""
+    words: List[str] = []
     for name in names:
-        normalized = fingerprint_name(name)
+        normalized = normalizer(name)
         if normalized is not None:
-            words.update(normalized.split(WS))
+            for word in normalized.split(WS):
+                if len(word) >= min_length:
+                    if processor is not None:
+                        word = processor(word)
+                    words.append(word)
     return words
 
 
-@lru_cache(maxsize=1024)
 def normalize_name(original: str) -> Optional[str]:
     """Normalize a legal entity name."""
-    return clean_strict(original)
+    return clean_name_ascii(original)
 
 
-@lru_cache(maxsize=512)
+def phonetic_token(token: str) -> str:
+    return metaphone_token(token)
+
+
+@lru_cache(maxsize=1024)
+def metaphone_token(token: str) -> str:
+    if token.isalpha() and len(token) > 1:
+        return metaphone(token)
+    return token.upper()
+
+
+@lru_cache(maxsize=1024)
+def soundex_token(token: str) -> str:
+    if token.isalpha() and len(token) > 1:
+        return soundex(token)
+    return token.upper()
+
+
+@lru_cache(maxsize=1024)
 def levenshtein(left: str, right: str) -> int:
     """Compute the Levenshtein distance between two strings."""
-    return Levenshtein.distance(left[:128], right[:128])
+    return damerau_levenshtein_distance(left[:128], right[:128])
+
+
+@lru_cache(maxsize=1024)
+def jaro_winkler(left: str, right: str) -> float:
+    score = jaro_winkler_similarity(left, right)
+    return score if score > 0.6 else 0.0
+
+
+def clean_identifier(
+    value: str, min_length: int = 6, max_length: int = 100
+) -> Optional[str]:
+    """Clean up an identifier for comparison."""
+    value = ID_CLEAN.sub("", value.upper())
+    if len(value) < min_length or len(value) > max_length:
+        return None
+    return value
+
+
+def list_intersection(left: List[str], right: List[str]) -> int:
+    """Return the number of elements in the intersection of two lists, accounting
+    properly for duplicates."""
+    overlap = 0
+    remainder = list(right)
+    for elem in left:
+        try:
+            remainder.remove(elem)
+            overlap += 1
+        except ValueError:
+            pass
+    return overlap
 
 
 def pack_prop(schema: str, prop: str) -> str:
